
namespace nanojit
{
	// Basics:
	// - 'entry' records the state of the native machine stack at particular
	//   points during assembly.  Each entry represents four bytes.
	//
	//
	// - The stack also holds spilled values, in which case each slot holding
	//   a spilled value (one slot for 32-bit values, two slots for 64-bit
	//   values) contains a pointer to the instruction defining the spilled
	//   value.
	//
	// - Each LIns has a "reservation" which includes a stack index,
	//   'arIndex'.  Combined with AR, it provides a two-way mapping between
	//   stack slots and LIR instructions.
	//
	// - Invariant: the two-way mapping between active stack slots and their
	//   defining/allocating instructions must hold in both directions and be
	//   unambiguous.  More specifically:
	//
	//   * An LIns can appear in at most one contiguous sequence of slots in
	//     AR, and the length of that sequence depends on the opcode (1 slot
	//     for instructions producing 32-bit values, 2 slots for instructions
	//     producing 64-bit values).
	//
	//   * An LIns named by 'entry[i]' must have an in-use reservation with
	//     arIndex==i (or an 'i' indexing the start of the same contiguous
	//     sequence that 'entry[i]' belongs to).
	//
	//   * And vice versa:  an LIns with an in-use reservation with arIndex==i
	//     must be named by 'entry[i]'.
	//
	//   * If an LIns's reservation names has arIndex==0 then LIns should not
	//     be in 'entry[]'.
	//
	class AR
	{
	private:
		unsigned int    _stackTop;						/* count of entries allocated */
		unsigned int    _firstEntryAllocationIndex;			/* starting index for stack allocation */
		LIns*           _entries[NANOJIT_MAX_STACK_ENTRY];	/* maps to contiguous locations relative to the frame pointer. */

		static LIns* const BAD_ENTRY;

		bool isEmptyRange(unsigned int start, unsigned int nStackSlots) const;

	public:
		static unsigned int nStackSlotsFor(LIns* ins);

	public:
		AR();

		unsigned int stackTop() const;
		unsigned int firstEntryAllocationIndex() const;

		void reset(unsigned int firstEntryIndex);
		void releaseEntryAt(unsigned int  i);
		unsigned int reserveEntry(LIns* ins); /* return 0 if unable to reserve the entry */
		bool reallocateEntry(LIns* ins, uint32_t i);

		#ifdef NANOJIT_DEBUG
		void validateQuick();
		void validateFull();
		void validate();
		bool isValidEntry(unsigned int idx, LIns* ins) const; /* return true iff idx and ins are matched */
		bool isValidEntry(unsigned int idx, LInsCompilationData* insCompilationData) const; /* return true iff idx and ins are matched */
		void checkForResourceConsistency(const RegAlloc& regs);
			#if defined(NANOJIT_VERBOSE)
			void checkForResourceLeaks(LInsPrinter* printer, LogControl* logc) const;
			#endif
		#endif

		class Iter
		{
		private:
			const AR& _ar;
			// '_i' points to the start of the entries for an LIns, or to the first NULL entry.
			uint32_t _i;
		public:
			inline Iter(const AR& ar) : _ar(ar), _i(0) { }
			bool next(LIns*& ins, unsigned int& nStackSlots, int& offset);             // get the next one (moves iterator forward)
		};
	};

	inline AR::AR()
	{
		 _entries[0] = BAD_ENTRY;
		 reset(0);
	}

	inline unsigned int AR::nStackSlotsFor(LIns* ins)
	{
		uint32_t n = 0;

		switch (ins->retType()) {
			case LTy_I:		n = ins->getArSize();	break;
			case LTy_F:		n = 4;					break;
			CASE64(LTy_Q:)
			case LTy_D:		n = 8;					break;
			case LTy_V:		NanoAssert(0);			break;
			default:		NanoAssert(0);			break;
		}
		return n;
	}

	inline unsigned int AR::stackTop() const {
		return _stackTop;
	}

	inline unsigned int AR::firstEntryAllocationIndex() const {
		return _firstEntryAllocationIndex;
	}

} // namespace nanojit


